#!/usr/bin/env python 
"""
Copyright (C) 2009 Nick Drobchenko, nick@cnc-club.ru
based on gcode.py (C) 2007 hugomatic... 
based on addnodes.py (C) 2005,2007 Aaron Spike, aaron@ekips.org
based on dots.py (C) 2005 Aaron Spike, aaron@ekips.org

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

###
###		Gcode tools v 1.0.0. 28.01.10.
###

import inkex, simplestyle, simplepath
import os
import math
import bezmisc
import re
import copy
import sys

biarc_style = {
		'biarc0':	simplestyle.formatStyle({ 'stroke': '#88f', 'fill': 'none', 'stroke-width':'1' }),
		'biarc1':	simplestyle.formatStyle({ 'stroke': '#8f8', 'fill': 'none', 'stroke-width':'1' }),
		'line':		simplestyle.formatStyle({ 'stroke': '#f88', 'fill': 'none', 'stroke-width':'1' }),
		'area':		simplestyle.formatStyle({ 'stroke': '#777', 'fill': 'none', 'stroke-width':'0.1' }),
	}

biarc_style_dark = {
		'biarc0':	simplestyle.formatStyle({ 'stroke': '#33a', 'fill': 'none', 'stroke-width':'1' }),
		'biarc1':	simplestyle.formatStyle({ 'stroke': '#3a3', 'fill': 'none', 'stroke-width':'1' }),
		'line':		simplestyle.formatStyle({ 'stroke': '#a33', 'fill': 'none', 'stroke-width':'1' }),
		'area':		simplestyle.formatStyle({ 'stroke': '#222', 'fill': 'none', 'stroke-width':'0.3' }),
	}

biarc_style_dark_area = {
		'biarc0':	simplestyle.formatStyle({ 'stroke': '#33a', 'fill': 'none', 'stroke-width':'0.1' }),
		'biarc1':	simplestyle.formatStyle({ 'stroke': '#3a3', 'fill': 'none', 'stroke-width':'0.1' }),
		'line':		simplestyle.formatStyle({ 'stroke': '#a33', 'fill': 'none', 'stroke-width':'0.1' }),
		'area':		simplestyle.formatStyle({ 'stroke': '#222', 'fill': 'none', 'stroke-width':'0.3' }),
	}

biarc_style_i = {
		'biarc0':	simplestyle.formatStyle({ 'stroke': '#880', 'fill': 'none', 'stroke-width':'1' }),
		'biarc1':	simplestyle.formatStyle({ 'stroke': '#808', 'fill': 'none', 'stroke-width':'1' }),
		'line':		simplestyle.formatStyle({ 'stroke': '#088', 'fill': 'none', 'stroke-width':'1' }),
		'area':		simplestyle.formatStyle({ 'stroke': '#999', 'fill': 'none', 'stroke-width':'0.3' }),
	}

biarc_style_dark_i = {
		'biarc0':	simplestyle.formatStyle({ 'stroke': '#dd5', 'fill': 'none', 'stroke-width':'1' }),
		'biarc1':	simplestyle.formatStyle({ 'stroke': '#d5d', 'fill': 'none', 'stroke-width':'1' }),
		'line':		simplestyle.formatStyle({ 'stroke': '#5dd', 'fill': 'none', 'stroke-width':'1' }),
		'area':		simplestyle.formatStyle({ 'stroke': '#aaa', 'fill': 'none', 'stroke-width':'0.3' }),
	}


math.pi2 = math.pi*2
straight_tolerance = 0.0001
straight_distance_tolerance = 0.0001
parse_codes_maxdepth = 4
#scalex = scalex*scale 
#scaley = scaley*scale 



defaults = {
		'header': '%\n(Generated by gcode_tools from inkscape.) \nM3 \n',
		'footer': 'M5 \nG00 X0.0000 Y0.0000 \nM2 \n(end)\n%'
}

###
###		Just simple output function for better debugging
###
#if os.path.isfile("/home/nick/output.txt") :os.remove("/home/nick/output.txt")
#def print_(s=''):
#	f = open("/home/nick/output.txt","a")
#	f.write(str(s))
#	f.write("\n")
#	f.close()





class P:
	def __init__(self, x, y=None):
		if not y==None:
			self.x, self.y = float(x), float(y)
		else:
			self.x, self.y = float(x[0]), float(x[1])

	def __add__(self, other):
		return P(self.x + other.x, self.y + other.y)

	def __sub__(self, other):
		return P(self.x - other.x, self.y - other.y)

	def __neg__(self):
		return P(-self.x, -self.y)


	def __mul__(self, other):
		if isinstance(other, P):
			return self.x * other.x + self.y * other.y
		return P(self.x * other, self.y * other)
	__rmul__ = __mul__

	def __div__(self, other):
		return P(self.x / other, self.y / other)

	def mag(self):
		return math.hypot(self.x, self.y)

	def unit(self):
		h = self.mag()
		if h: return self / h
		else: return P(0,0)
	
	def dot(self, other):
		return self.x * other.x + self.y * other.y

	def rot(self, theta):
		c = math.cos(theta)
		s = math.sin(theta)

		return P(self.x * c - self.y * s,  self.x * s + self.y * c)

	def angle(self):
		return math.atan2(self.y, self.x)

	def __repr__(self): return '%f,%f' % (self.x, self.y)
	def pr(self): return "%.2f,%.2f" % (self.x, self.y)
	def to_list(self): return [self.x, self.y]	





def tpoint((x1,y1), (x2,y2), t = 0.5):
    return [x1+t*(x2-x1),y1+t*(y2-y1)]

def cspbezsplit( bez, t=0.5):
	m1=tpoint(bez[0],bez[1],t)
	m2=tpoint(bez[1],bez[2],t)
	m3=tpoint(bez[2],bez[3],t)
	m4=tpoint(m1,m2,t)
	m5=tpoint(m2,m3,t)
	m=tpoint(m4,m5,t)
	return  [bez[0],m1,m4,m], [m,m5,m3,bez[3]] 

def cspbezsplitatlength(bez, l = 0.5, tolerance = 0.001):
    t = bezmisc.beziertatlength(bez, l, tolerance)
    return cspbezsplit(bez, t)	

class Gcode(inkex.Effect):

	def __init__(self):
		inkex.Effect.__init__(self)
		self.OptionParser.add_option("-d", "--directory",				action="store", type="string", 		dest="directory", default="/home/",				help="Directory for gcode file")
		self.OptionParser.add_option("-f", "--filename",				action="store", type="string", 		dest="file", default="-1.0",					help="File name")			
		self.OptionParser.add_option("-u", "--Xscale",					action="store", type="float", 		dest="Xscale", default="1.0",					help="Scale factor X")	
		self.OptionParser.add_option("-v", "--Yscale",					action="store", type="float", 		dest="Yscale", default="1.0",					help="Scale factor Y")
		self.OptionParser.add_option("",   "--Zscale",					action="store", type="float", 		dest="Zscale", default="1.0",					help="Scale factor Z")				
		self.OptionParser.add_option("-x", "--Xoffset",					action="store", type="float", 		dest="Xoffset", default="0.0",					help="Offset along X")	
		self.OptionParser.add_option("-y", "--Yoffset",					action="store", type="float", 		dest="Yoffset", default="0.0",					help="Offset along Y")
		self.OptionParser.add_option("",   "--Zoffset",					action="store", type="float", 		dest="Zoffset", default="0.0",					help="Offset along Z")
		self.OptionParser.add_option("-s", "--Zsafe",					action="store", type="float", 		dest="Zsafe", default="0.5",					help="Z above all obstacles")
		self.OptionParser.add_option("-z", "--Zsurface",				action="store", type="float", 		dest="Zsurface", default="0.0",					help="Z of the surface")
		self.OptionParser.add_option("-c", "--Zdepth",					action="store", type="float", 		dest="Zdepth", default="-0.125",					help="Z depth of cut")
		self.OptionParser.add_option("",   "--Zstep",					action="store", type="float", 		dest="Zstep", default="-0.125",					help="Z step of cutting")		
		self.OptionParser.add_option("-p", "--feed",					action="store", type="float", 		dest="feed", default="4.0",						help="Feed rate in unit/min")
		self.OptionParser.add_option("",   "--cut-area",				action="store", type="string", 		dest="cut_area", default="false",				help="Cut internal area of the curve")				
		self.OptionParser.add_option("",   "--tool-diameter",			action="store", type="float", 		dest="tool_diameter", default="3",				help="Tool diameter used for area cutting")		
		self.OptionParser.add_option("",   "--max-area-curves",			action="store", type="int", 		dest="max_area_curves", default="100",			help="Maximum area curves for each area")
		self.OptionParser.add_option("",   "--area-inkscape-radius",	action="store", type="int", 		dest="area_inkscape_radius", default="-10",		help="Radius for preparing curves using inkscape")
		self.OptionParser.add_option("",   "--unit",					action="store", type="string", 		dest="unit", default="G21 (All units in mm)\n",	help="Units")
		self.OptionParser.add_option("",   "--function",				action="store", type="string", 		dest="function", default="Curve",				help="What to do: Curve|Area|Area inkscape")
		self.OptionParser.add_option("",   "--tab",						action="store", type="string", 		dest="tab", default="",							help="Tab")
		self.OptionParser.add_option("",   "--generate_not_parametric_code",action="store", type="inkbool",	dest="generate_not_parametric_code", default=False,help="Generated code will be not parametric.")		

# Additional options
		self.OptionParser.add_option("",   "--min-arc-radius",			action="store", type="float", 		dest="min_arc_radius", default=".1",				help="All arc having radius less than minimum will be considered as straight line")		

	def draw_line(self, l, group=None, style=biarc_style):
		if group==None:
			group = inkex.etree.SubElement( self.biarcGroup, inkex.addNS('g','svg') )
		s = "M %s,%s " % (l[0][0],l[0][1])
		for i in range(1,len(l)):
			s += "L %s,%s " % (l[i][0],l[i][1])
		inkex.etree.SubElement(	group, inkex.addNS('path','svg'), 
				{
					'style': style['area'],
					'd':s,
				})

			

	def biarc(self, bez):
		P0, P4 = P(bez[0]), P(bez[3])
		TS, TE, v = (P(bez[1])-P0), -(P(bez[2])-P4), P0 - P4
		tsa, tea, va = TS.angle(), TE.angle(), v.angle()
		if TE.mag()<straight_distance_tolerance and TS.mag()<straight_distance_tolerance:	
			# Both tangents are zerro - line straight
			return ['Add L']
		if TE.mag() < straight_distance_tolerance:
			TE = -(TS+v).unit()
			r = TS.mag()/v.mag()*2
		elif TS.mag() < straight_distance_tolerance:
			TS = -(TE+v).unit()
			r = 1/( TE.mag()/v.mag()*2 )
		else:	
			r=(P(bez[1])-P0).mag()/(P(bez[2])-P4).mag()
		TS, TE = TS.unit(), TE.unit()
		tang_are_parallel = ((tsa-tea)%math.pi<straight_tolerance or math.pi-(tsa-tea)%math.pi<straight_tolerance )
		if tang_are_parallel  and v.mag()<straight_distance_tolerance or TE.mag()<straight_distance_tolerance or TS.mag()<straight_distance_tolerance:
			# Both tangents are parallel and start and end are the same - line straight
			# or one of tangents still smaller then tollerance
			return ['Add L']
		if tang_are_parallel and  1-abs(TS*v/(TS.mag()*v.mag()))<straight_tolerance :
			# Both tangents and v are parallel - line straight
			return ['Add L']  				
		
		c,b,a = v*v, 2*v*(r*TS+TE), 2*r*(TS*TE-1)

		if v.mag()==0 and abs(TE.angle())!=abs(TS.angle()):
			return ['Need split']
		asmall, bsmall, csmall = abs(a)<10**-10,abs(b)<10**-10,abs(c)<10**-10 
		if 		asmall and b!=0:	beta = -c/b
		elif 	csmall and a!=0:	beta = -b/a 
		elif not asmall:	 
			discr = b*b-4*a*c
			if discr < 0:	raise ValueError, (a,b,c,discr)
			disq = discr**.5
			beta1 = (-b - disq) / 2 / a
			beta2 = (-b + disq) / 2 / a
			if beta1*beta2 > 0 :	raise ValueError, (a,b,c,disq,beta1,beta2)
			beta = max(beta1, beta2)
		elif	asmall and bsmall:	return ['Need split']
		alpha = beta * r
		ab = alpha + beta 
		P1 = P0 + alpha * TS
		P3 = P4 - beta * TE
		P2 = (beta / ab)  * P1 + (alpha / ab) * P3
		
### Draw line for testing
#		self.draw_line([[P0.x,P0.y], [P1.x,P1.y], [P2.x,P2.y], [P3.x,P3.y], [P4.x,P4.y]])

		return ['Biarc', [ [P0,P1,P2,P3,P4], alpha, beta ] ]



	
	def parse_code(self, cmd, params, cpos, depth=1):
		if cmd == 'M':
			self.current_curve_moved_from = params[-2:]
			return params[-2:], [[cpos, 'move', 0, 0]]
		if cmd == 'C':
			bez = ( cpos, params[:2], params[2:4], params[-2:] )
			b = self.biarc(bez)
			if b[0] == "Add L":
				cmd = 'L'
			elif b[0] == 'Need split':
				if depth>parse_codes_maxdepth:
					cmd = 'L'
				else:	
					p=[0]*6
					bez1, bez2 = cspbezsplitatlength(bez)
					(p[0], p[1]), (p[2], p[3]), (p[4], p[5]) = bez1[1:]
					cpos, curve = self.parse_code('C',p,cpos, depth+1)
					(p[0], p[1]), (p[2], p[3]), (p[4], p[5]) = bez2[1:]
					cpos, curve1 = self.parse_code('C',p,cpos, depth+1)
					curve.extend(curve1)
					return cpos, curve
			elif b[0] == 'Biarc':
				def calculate_arc(P0,P1,P2):
					D = (P0+P2)/2
					if (D-P1).mag()==0: return [ [P0.x,P0.y], 'line', 0, 0, 0]
					R = D - ( (D-P0).mag()**2/(D-P1).mag() )*(P1-D).unit()
					p0a, p1a, p2a = (P0-R).angle()%(2*math.pi), (P1-R).angle()%(2*math.pi), (P2-R).angle()%(2*math.pi)
					alpha =  (p2a - p0a) % (2*math.pi)					
					if (p0a<p2a and  (p1a<p0a or p2a<p1a))	or	(p2a<p1a<p0a) : 
						alpha = -2*math.pi+alpha
					if abs(R.x)>1000000 or abs(R.y)>1000000  or (R-P0).mag<self.options.min_arc_radius :
						
						return [[P0.x,P0.y], 'line', 0, 0, 0] 
					else :	
						return  [[P0.x,P0.y], 'arc', [R.x,R.y], alpha, 0] 
					
				return [b[1][0][4].x,b[1][0][4].y], [calculate_arc(b[1][0][0], b[1][0][1], b[1][0][2]), calculate_arc(b[1][0][2], b[1][0][3], b[1][0][4])]
				
		if cmd == 'Z': 
			if (P(cpos) - P(self.current_curve_moved_from)).mag() > straight_distance_tolerance :
				return self.current_curve_moved_from, [[cpos, 'line',0 , 0, 0], [self.current_curve_moved_from, 'end', 'close']]
			else : return self.current_curve_moved_from, [ [self.current_curve_moved_from, 'end', 'close'] ]
			
		if cmd == 'L':
			return params[-2:], [[cpos, 'line',0 , 0, 0]]
		if cmd == 'A':
			r = (params[0]+params[1])/2
			sign, P2 = 1 if params[4]==params[3] else -1, P(params[5], params[6])
			P0 = P(cpos)
			RN = P2-P0
			if r**2 - RN.mag()**2/4<0 : rm = 0 
			else : rm = math.sqrt(r**2 - RN.mag()**2/4)
			RN = RN.unit()
			R = P(RN.y,-RN.x)
			R = (P2+P0)/2 + sign*R*rm
			p0a, p2a = (P0-R).angle()%(2*math.pi), (P2-R).angle()%(2*math.pi)
			alpha =  (p2a - p0a) % (2*math.pi)					
			if sign>0: 
				alpha = -2*math.pi+alpha
			
			
			return [P2.x,P2.y], [	[cpos, 'arc', [R.x,R.y], alpha, 0]	]
			
		raise ValueError, (cmd,params)


	def draw_curve(self, curve, group=None, style=biarc_style):
###		try:
###			t = node.get('transform')
###			self.group.set('transform', t)
###		except:
###			pass
		if group==None:
			group = inkex.etree.SubElement( self.biarcGroup, inkex.addNS('g','svg') )
		s, arcn = '', 0
		for si in curve:
			if s!='':
				if s[1] == 'line':
					inkex.etree.SubElement(	group, inkex.addNS('path','svg'), 
							{
								'style': style['line'],
								'd':'M %s,%s L %s,%s' % (s[0][0], s[0][1], si[0][0], si[0][1]),
								'comment': str(s)
							}
						)
				elif s[1] == 'arc':
					arcn += 1
					sp = s[0]
					c = s[2]
					a =  ( (P(si[0])-P(c)).angle() - (P(s[0])-P(c)).angle() )%math.pi2 #s[3]
					if s[3]*a<0: 
							if a>0:	a = a-math.pi2
							else: a = math.pi2+a
					r = math.sqrt( (sp[0]-c[0])**2 + (sp[1]-c[1])**2 )
					a_st = ( math.atan2(sp[0]-c[0],- (sp[1]-c[1])) - math.pi/2 ) % (math.pi*2)
					if a>0:
						a_end = a_st+a
					else: 
						a_end = a_st*1
						a_st = a_st+a	
					inkex.etree.SubElement(	group, inkex.addNS('path','svg'), 
						 {
							'style': style['biarc%s' % (arcn%2)],
							 inkex.addNS('cx','sodipodi'):		str(c[0]),
							 inkex.addNS('cy','sodipodi'):		str(c[1]),
							 inkex.addNS('rx','sodipodi'):		str(r),
							 inkex.addNS('ry','sodipodi'):		str(r),
							 inkex.addNS('start','sodipodi'):	str(a_st),
							 inkex.addNS('end','sodipodi'):		str(a_end),
							 inkex.addNS('open','sodipodi'):	'true',
							 inkex.addNS('type','sodipodi'):	'arc',
							'comment': str(s)
						})
			s = si
					
		
	

	def offset(self, s, si, d):
		s = copy.copy(s)
		if s[1] == 'line':
			sp = P(s[0])
			n = P(s[0][1] - si[0][1], si[0][0] - s[0][0]).unit()*d
			s[0] = [s[0][0] + n.x, s[0][1]+n.y]
			s[4] = [si[0][0] + n.x, si[0][1]+n.y]
		if s[1] == 'arc':
			n = P(s[2][0] - s[0][0], s[2][1] - s[0][1]).unit()*d
			if s[3] < 0 :
				n = -n  
			s[0] = [s[0][0] + n.x, s[0][1]+n.y]
			n = P(s[2][0] - si[0][0], s[2][1] - si[0][1]).unit()*d
			if s[3] < 0 :
				n = -n  
			s[4] = [si[0][0] + n.x, si[0][1]+n.y]
		return s
		
		
		
	def find_true_intersections(self, s,si):
		intersect = self.find_intersections(s, si)
		res = []
		for i in intersect:
			ip, a = self.get_ip_type(s,i)
			ipi, ai = self.get_ip_type(si,i)
			if ip==ipi=='t': 
				res += [ [a,i] ]
		return res
		
		
	def find_intersections(self, s, si):
		if s[1]==si[1]:
			if s[1] == 'line':				
				sx, sy = s[4][0]-s[0][0], s[4][1]-s[0][1]
				six, siy = si[4][0]-si[0][0], si[4][1]-si[0][1]
				if (sx==0 and sy==0) or (six==0 and siy==0) : return []
				if sx==six==0 :
					if sy != siy : return []
					else : return [ [ (s[4][0]+si[0][0])/2,(s[4][1]+si[0][1])/2 ] ]  # Overlap
				if ( abs(six*sy-sx*siy)<straight_tolerance ): 
					if ( sy*(si[0][0] - s[0][0]) == sx*(si[0][1]-s[0][1]) ) : return [ [(s[4][0]+si[0][0])/2,(s[4][1]+si[0][1])/2] ] # Overlap
					else : return []
				if abs(sx) < straight_tolerance :
					x = (s[0][0]+s[4][0])/2
					if abs(six) < straight_tolerance : y = (si[0][1]+si[4][1])/2
					else : y = x * siy/six - si[0][0]*siy/six + si[0][1]
				elif abs(six) < straight_tolerance :
					x = (si[0][0]+si[4][0])/2
					if abs(sx) < straight_tolerance : y = (s[0][1]+s[4][1])/2
					else : y = x * sy/sx - s[0][0]*sy/sx + s[0][1]
				else:	
					a, ai = sy/sx, siy/six
					b, bi = - s[0][0]*sy/sx + s[0][1], - si[0][0]*siy/six + si[0][1]
					if bi-b != 0:						
						x = (bi-b)/(a-ai)
					else: x = 0	
					y = a*x+b
				return [[x,y]]
			elif s[1]== 'arc' :
				s = copy.copy(s)
				s[1] = 'line_abc'
				r2, ri2 = (s[0][0]-s[2][0])**2 + (s[0][1]-s[2][1])**2, (si[0][0]-si[2][0])**2 + (si[0][1]-si[2][1])**2
				x1,y1 = si[2][0], si[2][1]
				x2,y2 = s[2][0] - x1, s[2][1] - y1
				A = -2*(x2)
				B = -2*(y2)
				C = (x2)**2+(y2)**2 - r2 + ri2 - A*x1-B*y1
				s[0] = [A,B,C]
			
		if si[1]!=s[1] and si[1] in ['line','line_abc','arc'] and s[1] in ['line','line_abc','arc'] :
			
			if si[1] == 'arc' : a, l = si, s
			else :	a, l = s, si
			if l[1]=='line_abc': 
				A,B,C = l[0]
				C += A*a[2][0] + B*a[2][1]
			else :
				A, B = l[4][1] - l[0][1], - ( l[4][0] - l[0][0])
				C =  - l[0][0]*A - l[0][1]*B  + A*a[2][0] + B*a[2][1]
			r2 = (a[0][0]-a[2][0])**2 + (a[0][1]-a[2][1])**2
			if (A**2+B**2)>0:
				x0, y0 = -A*C/(A**2+B**2), -B*C/(A**2+B**2)
			else : return []	
			if 	abs(A*x0+B*y0+C)>abs(-A*x0-B*y0+C) : x0, y0 = -x0,-y0
			if abs (C**2 - r2*(A**2+B**2)) < straight_distance_tolerance: 
				if abs(x0**2+y0**2-r2) < straight_distance_tolerance:
					return [[x0+a[2][0], y0+a[2][1]]]
				else: return []	
			elif C**2 > r2*(A**2+B**2) : 
				return []
			else :
				m = math.sqrt( (r2 - C**2/(A**2+B**2)) / (A**2+B**2) )
				return [[a[2][0]+x0-B*m, a[2][1]+y0+A*m], [a[2][0]+x0+B*m, a[2][1]+y0-A*m]]
				
		else :	
			return []

	def between(self, c,x,y):
		return x-straight_tolerance<=c<=y+straight_tolerance or y-straight_tolerance<=c<=x+straight_tolerance
	def between_ang(self, c1,a1,b1):
		a,b,c = a1%math.pi2,b1%math.pi2,c1%math.pi2
		if a<=b:
			return a<=c<=b
		else:
			return not b<=c<=a
	
	def get_ip_type(self, s, int_):
		if s[1] == 'line':
			x1,y1,x2,y2 = s[0][0], s[0][1], s[4][0], s[4][1]
			a = math.sqrt( (x1-int_[0])**2 + (y1-int_[1])**2)
			if self.between(int_[0],x1,x2) and self.between(int_[1],y1,y2): return 't', a
			else:
				if abs(x1-x2)>abs(y1-y2) :
					if self.between(x2,x1,int_[0]): return 'fp', a	
					else: return 'fn', -a
				else:
					if self.between(y2,y1,int_[1]): return 'fp', a	
					else: return 'fn', -a

		elif s[1] == 'arc':
			P0, R, Pi = P(s[0]), P(s[2]), P(int_)
			alpha = ((Pi-R).angle() - (P0-R).angle())
			if s[3]*alpha<0: 
				if alpha>0:	alpha = alpha-math.pi2
				else: alpha = math.pi2+alpha
			if self.between(alpha,0,s[3]) or min(abs(alpha),abs(alpha-s[3]))<straight_tolerance : return 't', alpha
			else : return 'f', alpha
	
	def angle(self,s,e,c, direction='cw'):
		ang = (P(e)-P(c)).angle() - (P(s)-P(c)).angle()
		if direction =='cw': return ang % math.pi2
		elif direction =='ccw': return ang % math.pi2 - math.pi2
		elif direction =='min': 
			if abs(ang % math.pi2 - math.pi2) < ang % math.pi2 :
				return ang % math.pi2 - math.pi2
			else: return ang % math.pi2
	
	
	def add_arc(self, s,si,c,direction = 1):
		ang = self.angle(s[4],si[0],c,'min')
		c = P(c)
		ps = P(s[4])-c
		pe = P(si[0])-c
		if (pe-ps).mag()<straight_tolerance : return None  
		return [ [(ps+c).x,(ps+c).y], 'arc', [c.x,c.y], ang,  [(pe+c).x,(pe+c).y] ]


	def offset_join(self,s,si, curve):
				int_ = self.find_intersections(s,si)
				if len(int_)>1:	
					for k in range(len(int_)):
						int_[k] = [ (curve[0][0]-int_[k][0])**2 + (curve[0][1]-int_[k][1])**2, int_[k]]
					int_.sort()
					int_ = int_[0][1]
				elif int_ != [] : int_ = int_[0]
				ip={}
				if int_ != []:
					ip['s'],alpha = self.get_ip_type(s,int_)
					ip['si'],alphai = self.get_ip_type(si,int_)
						
				if s[1]==si[1]=='line':
					if ip =={} or ip['s']=='fp' and ip['si']=='fn':
						a = self.add_arc(s,si,curve[0])
						if a!=None : return [ a, si ], None 
						else : return [ [s[4]]+si[1:] ], None 
					elif ip['s']==ip['si']=='t':
						return [ ([int_]+si[1:]) ], [ s[0], s[1],s[2],s[3],int_ ]
					elif ip['s']=='fp' and ip['si']=='fp' or ip['s']!='t' and ip['si']=='t' or ip['s']=='t' and ip['si']!='t':
						return [  [ s[4],'line',0 ,0 , si[0] ] ,si ], None
					else :
						return [  [ s[4],'line',0 ,0 , si[0] ] ,si ], None
				elif s[1]==si[1]=='arc':
					if ip=={} or ip['s']=='f' and ip['si']=='f' or ip['s']!=ip['si']:
						a = self.add_arc(s,si,curve[0])
						if a!=None : return [ a, si ], None 
						else : return [ [s[4]]+si[1:] ], None 
					elif ip['s']=='t' and ip['si']=='t':
						return [ [int_, 'arc', si[2], si[3]-alphai, si[4]] ], [s[0], s[1],s[2], alpha, int_]
				elif s[1]=='arc' and si[1]=='line':
					if ip=={} or ip['s']!='t' and ip['si']!='t':
						a = self.add_arc(s,si,curve[0])
						if a!=None : return [ a, si ], None 
						else : return [ [s[4]]+si[1:] ], None 
					elif ip['s']=='t' and ip['si']=='t':
						return [ [int_, 'line', si[2], si[3], si[4]] ], [s[0], s[1],s[2], alpha, int_] 
					elif ip['s']!='t' and ip['si']=='t' or ip['s']=='t' and ip['si']!='t':
						return [  [ s[4],'line',0 ,0 , si[0] ] ,si ], None
				elif s[1]=='line' and si[1]=='arc':
					if ip=={} or ip['s']!='t' and ip['si']!='t':
						a = self.add_arc(s,si,curve[0])
						if a!=None : return [ a, si ], None 
						else : return [ [s[4]]+si[1:] ], None 
					elif ip['s']!='t' and ip['si']=='t' or ip['s']=='t' and ip['si']!='t':
						return[  [ s[4],'line',0 ,0 , si[0] ] ,si ], None
					elif ip['s']=='t' and ip['si']=='t':
						return [ [int_, 'arc', si[2], si[3]-alphai, si[4]] ], None
				else : 	return [ si ], None

		

	def offset_curve(self,curve,d):
		p1 = []		
		first_curve = None
		for i in range(len(curve)):
			si = curve[i]
			if si[1] == 'move':
				first_curve = None
				if len(p1)>0 and p1[-1][1]!='end':
					p1.extend([ [p1[-1][0],'end'] ])
				p1.extend([ si ])			
				continue	
			if si[1] == 'end': 
				if len(si)>2 and si[2] == 'close' :
						s = p1[-1]
						si = p1[first_curve[0]]
						cur, ns = self.offset_join(s,si,curve[first_curve[1]])	
						if ns != None : p1[-1] = ns
						p1.extend( cur[:-1] + [ [cur[-1][0], 'end', 'close'] ] )
						p1[first_curve[0]] = cur[-1]
						continue
				else :
					p1.extend([  [ p1[-1][4], 'end'  ]  ])
					continue
		
			if i==0 :
				p1.extend([self.offset(curve[i],curve[i+1],d)])
				continue
			s = p1[-1]
			
			if ( si[1] == 'arc' and  (d>0 and si[3]>0 and (P(si[0])-P(si[2])).mag() < d) or 
							(d<0 and si[3]<0 and (P(si[0])-P(si[2])).mag() < -d) ): continue
			si = self.offset(curve[i],curve[i+1],d)
			if si ==[]: continue
			if s[1] in ['line','arc'] and si[1] in ['line','arc'] and (P(s[4])-P(si[0])).mag<straight_distance_tolerance : continue
			
			cur, ns = self.offset_join(s,si,curve[i])	
			if ns != None : p1[-1] = ns
			if cur[0][1] in ['arc','line'] and first_curve == None: first_curve = [len(p1), i]
			p1.extend( cur )			

		return p1

		
		
	###
	###		Curve structure	
	###		[	[s - start point, type - line|arc|end|move, c - center, a - alpha [-2pi,2pi] illustrates direction, e - end point (optional)] ...	]
	###
	def parse_curve(self, d):
		p = simplepath.parsePath(d)
		cpos = [0,0]
		curve = []
		curves = []
		for cmd,params in p:
			cpos, c = self.parse_code(cmd,params,cpos, depth=1)
			if c[-1][1] == 'move': 
				if len(curve)>0 and curve[-1][1]!='end':
					if (P(last_moved_to)-P(c[-1][0])).mag()< straight_distance_tolerance :
						curve.extend([ [last_moved_to, 'end', 'close']  ])
					else :
						curve.extend([ [c[-1][0], 'end']  ])
				last_moved_to = c[-1][0]
			curve.extend(c)
		if curve!=[]: 		
			if curve[-1][1] != 'end':
					if (P(last_moved_to)-P(cpos)).mag()< straight_distance_tolerance :
						curve.extend([ [last_moved_to, 'end', 'close']  ])
					else :
						curve.extend([ [cpos, 'end']  ])
		return curve
			


	def generate_gcode(self, curve, depth):
		def coord(c):
			if self.options.generate_not_parametric_code:
				return c[0]*self.options.Xscale+self.options.Xoffset , -c[1]*self.options.Yscale+self.options.Yoffset
			else:
				return c[0], -c[1]
		np = self.options.generate_not_parametric_code
		g = ''
		s = []
		lg = 'G00'
		for si in curve:
			if s == []:	s = si; continue
			if s[1]	== 'move':
				if np:
					g += 'G00 Z%f'%(self.options.Zsafe*self.options.Zscale+self.options.Zoffset) + 'G00 X%f Y%f'%coord(si[0]) + 'G01 Z%f F%f\n'%(depth*self.options.Zscale+self.options.Zoffset,self.options.feed)
				else: 
					g += 'G00 Z[#11*#7 + #10] \n' +  'G00 X[%f*#5+#8] Y[%f*#6+#9] \n' % coord(si[0]) + 'G01 Z[%f*#7+#10] F#4\n' % (depth)
				lg = 'G00'
			elif s[1] == 'end':
				if np:
					g += 'G00 Z%f'%(self.options.Zsafe*self.options.Zscale+self.options.Zoffset) + 'G00 X%f Y%f'%coord(si[0]) + 'G01 Z%f F%f\n'%(depth*self.options.Zscale+self.options.Zoffset,self.options.feed)
				else:
					g += 'G00 Z[#11*#7 + #10] \n' +  'G00 X[%f*#5+#8] Y[%f*#6+#9] \n' % coord(si[0]) + 'G01 Z[%f*#7+#10] F#4\n' % (depth)
				lg = 'G00'
			elif s[1] == 'line':
				if np:
					g += 'G00 X%f Y%f'%coord(si[0]) + ('F%f'%(self.options.feed) if lg not in ['G01','G02','G03'] else '') +'\n'
				else:
					g += 'G01 X[%f*#5+#8] Y[%f*#6+#9] '	% coord(si[0]) + ('F #4' if lg not in ['G01','G02','G03'] else '') +'\n'
				lg = 'G01'
			elif s[1] == 'arc':
				r=coord([  (s[2][0]-s[0][0]), (s[2][1]-s[0][1])  ])
				if r[0]**2*self.options.Xscale+r[1]**2*self.options.Yscale>self.options.min_arc_radius:
					r1, r2 = (P(s[0])-P(s[2])), (P(si[0])-P(s[2]))
					if abs(r1.mag()-r2.mag()) < 0.001 :
						if np:
							g += ('G02' if s[3]>0 else 'G03')+ ' X%f Y%f '	% coord(si[0]) + ' I%f J%f ' % coord([  (s[2][0]-s[0][0]), (s[2][1]-s[0][1])  ]) + ('F%f'%self.options.feed if lg not in ['G01','G02','G03'] else '') +'\n'
						else:
							g += ('G02' if s[3]>0 else 'G03')+ ' X[%f*#5+#8] Y[%f*#6+#9] '	% coord(si[0]) + ' I[%f*#5+#8] J[%f*#6+#9] ' % coord([  (s[2][0]-s[0][0]), (s[2][1]-s[0][1])  ]) + ('F #4' if lg not in ['G01','G02','G03'] else '') +'\n'
					else:
						r = (r1.mag()+r2.mag())/2
						if np:
							g += ('G02' if s[3]>0 else 'G03')+ ' X%f Y%f '	% coord(si[0]) + ' R%f' % r*self.options.Xscale + ('F%f'%self.options.feed if lg not in ['G01','G02','G03'] else '') +'\n'						
						else:
							g += ('G02' if s[3]>0 else 'G03')+ ' X[%f*#5+#8] Y[%f*#6+#9] '	% coord(si[0]) + ' R[%f*#5]' % r + ('F #4' if lg not in ['G01','G02','G03'] else '') +'\n'						
								
					lg = 'G02'
				else:
					if np:
						g += 'G01 X%f Y%f '	% coord(si[0]) + ('F%f'%(self.options.feed) if lg not in ['G01','G02','G03'] else '') +'\n'
					else:
						g += 'G01 X[%f*#5+#8] Y[%f*#6+#9] '	% coord(si[0]) + ('F #4' if lg not in ['G01','G02','G03'] else '') +'\n'
					lg = 'G01'
			s = si
		if s[1] == 'end':
			if np:
				g += 'G00 Z%f \n'%(self.options.Zsafe*self.options.Zscale+self.options.Zoffset)
			else:
				g += 'G00 Z[#11*#7 + #10] \n'
			lg = 'G00'
		return g
		
		
	
	def prepare_curve(self,curve):
		pos = curve[0][0]
		c, s, sa = [], [], []
		for si in curve:
			if s == []: 
				s = si
				continue
			if s[1] == 'line':
				s[4] = si[0]
				l = math.sqrt( (s[0][0]-s[4][0])**2 + (s[0][1]-s[4][1])**2 )
			if s[1] == 'arc':
				s[4] = si[0]
				P0, R, P1 = P(s[0]), P(s[2]), P(s[4])
				alpha = ((P1-R).angle() - (P0-R).angle())
				if s[3]*alpha<0: 
					if alpha>0:	alpha = alpha-math.pi2
					else: alpha = math.pi2+alpha
				s[3] = alpha
				l =  abs(alpha)* (P0-R).mag() 
			if s[1] in ['move','end'] or l>straight_tolerance :
				if sa !=[] :
					sa[4] = s[0]
					c += [sa] + [s]
					sa = []
				else :
					c +=  [s]
			else :
				sa = [s[0], 'line',0,0,0] if sa == [] else sa				
			s = si
		c += [ si ]
		return c






		
	
	
	def clip(self,p1,p2, curve, d):
		
		p1, p2, curve = self.prepare_curve(p1), self.prepare_curve(p2), self.prepare_curve(curve)
		
		end_points = []
		for s in curve:	
			if s[1] == 'move' or s[1] == 'end' and (len(s)<3 or s[2] != 'close'):
				end_points += [s[0]]

		
		int_ = [[] for i in range(len(p1))]
		for i in range(len(p1)) :
			s = p1[i]
			if s[1] in ['move','end'] : continue
			intersect = []
			# Self intersections backward and forward
			startpos, follow = P(s[0]), False
			for j in range(i-1,0-1,-1):
				si = p1[j]
				if si[1] in ['line', 'arc'] and  ((startpos-P(si[4])).mag() > straight_distance_tolerance or follow): 
					follow = True
					int_[i] += self.find_true_intersections(s,si)

			startpos, follow = P(s[4]), False						
			for j in range(i+1,len(p1)):
				si = p1[j]
				if si[1] in ['line', 'arc'] and  ((startpos-P(si[0])).mag() > straight_distance_tolerance or follow): 
					follow = True
					int_[i] += self.find_true_intersections(s,si)

			
			# Intersections with inset
			for si in p2 :
				int_[i] += self.find_true_intersections(s,si)
			# Split by ending circle	
			for c in end_points :
				intersect = self.find_intersections(s,[ [c[0]+d,c[1]], 'arc', c, 0, 0, [c[0],c[1]+d] ])
			for inter in intersect:	
				ip, a = self.get_ip_type(s, inter)
				if ip == 't' :
					int_[i] += [ [a,inter] ]

		pa, pr = [], []
		for i in range(len(p1)):
			s = p1[i]
			if s[1] == 'move':
				if pa!=[]:
					if pa[-1][1] != 'end':
						pa += [  [pa[-1][4] if len(pa[-1])>4 else pa[-1][0] ,'end']  ]
					pr += [ pa ]
				pa = [ s ]
			elif s[1] == 'end' : 
				pa += [ s ]
				pr += [ pa ]
				pa = []
			elif int_[i]!=[]:
				int_[i].sort()
				if int_[i][len(int_[i])-1][0]<0 : int_[i].reverse()
				for j in range(len(int_[i])):
					point = int_[i][j]
					a = point[0] if j==0 else point[0]-int_[i][j-1][0]
					pa +=	[
							 	[s[0] if j==0 else int_[i][j-1][1],
							 	s[1] if s[1]!='arc' or abs(a)>straight_tolerance else 'line'  ,s[2],
							 	a,
							 	point[1]],
								[point[1], 'end']
							]
					pr += [ pa ]
					pa =	[	[point[1],'move']	]
				pa += [ [point[1], s[1], s[2], s[3]-point[0], s[4]] ]
			else : 
				pa += [ s ]
			
			
			
			###	
		
		result = []
		for p in pr:
			add = True
			for s in p:
				if s[1] not in ['line','arc']: continue
				if s[1] == 'line' :
					mp = [ (s[0][0]+s[4][0])/2,  (s[0][1]+s[4][1])/2 ] 
				elif s[1] == 'arc' :	
					R,P0,P1 = P(s[2]),P(s[0]), P(s[4])
					alpha = ((P1-R).angle() - (P0-R).angle())/2
					if s[3]*alpha<0: 
						if alpha>0:	alpha = alpha-math.pi
						else: alpha = math.pi+alpha
					mp = (P0-R).rot(alpha) + R
					mp = [mp.x, mp.y]
				dist,x,y = self.get_distance_to_curve(mp, curve)
				if dist < abs(d)-straight_distance_tolerance*10 :
#					try:				inkex.etree.SubElement(	self.biarcGroup, inkex.addNS('path','svg'), {'style': biarc_style_i['line'],
#						'd':'M %f,%f L %f,%f' % (mp[0],mp[1],x,y)} )
#					except: pass						 
					add=False
					break
			if add: result += [ p ]

		res = []
		for c in result :
			res = res +  c 
		return res
			
			
			
	def get_distance_to_curve(self, p, curve):
		dist = None
		p = P(p)
		xd,yd = 0, 0
		for s in curve:
			d = None
			if s[1] == 'line':
				A, B = s[4][1] - s[0][1], - ( s[4][0] - s[0][0])
				C =  - s[0][0]*A - s[0][1]*B  + A*p.x + B*p.y
				if (A**2+B**2)>0:
					i = P(-A*C/(A**2+B**2) + p.x, -B*C/(A**2+B**2) + p.y)
					if self.between(i.x,s[0][0],s[4][0]) and self.between(i.y,s[0][1],s[4][1]):
						x,y = i.x,i.y				
						d = (p-i).mag()
					else:	
						d1, d2 = (P(s[0])-p).mag(), (P(s[4])-p).mag()
						if d1<d2 : d, x,y  = d1, s[0][0], s[0][1] 
						else : d, x,y  = d2, s[4][0], s[4][1] 
			elif s[1] == 'arc':
				c = P(s[2])
				r = (P(s[0])-c).mag()
				if r>0 :
					i = c + (p-c).unit()*r
					ip, a = self.get_ip_type(s, [i.x,i.y])
					if ip == 't' : 
						d = (p-i).mag()
						x,y = i.x,i.y
					else : 
						d1, d2 = (p-P(s[0])).mag(), (p-P(s[4])).mag()
						if d1<d2 : d, x,y  = d1, s[0][0], s[0][1] 
						else : d, x,y  = d2, s[4][0], s[4][1] 
				else :
					d, x,y = (P(p) - P(s[2])).mag(), s[2][0], s[2][1]
				
			if dist!=None : 
				if d !=None and dist>d:
					dist, xd,yd = d, x,y
			elif d!=None :
					dist, xd,yd = d, x,y
		return dist, xd,yd	

	def curves_to_array(curve):
		pr = []
		pa = []
		for s in curve:
			if s[1] == 'end':
				pa.extend([ s ])
				pr.extend([ pa ])
				pa = []
			elif s[1] == 'move':
				if pa!=[]:
					if pa[-1][1] != 'end':
						pa.extend( [pa[-1][0],'end'] ) 
					pr.extend( [pa] )
				pa = [ s ]
			else: pa.extend( [s] )
		return pr	
	
	def get_bounds(self,curve):
		bounds = []
		for s in curve:
			bound = []
			if s[1] == 'line': 
				bound = [min(s[0][0],s[4][0]), min(s[0][1],s[4][1]), max(s[0][0],s[4][0]), max(s[0][1],s[4][1])] 
			elif s[1] == 'arc':
				left = math.pi
				right = 0
				bottom = - math.pi/2
				top = math.pi/2
				if  s[3]>0 : 
					st,end = (P(s[0])-P(s[2])).angle(), (P(s[4])-P(s[2])).angle()
				else :
					end,st = (P(s[0])-P(s[2])).angle(), (P(s[4])-P(s[2])).angle()
				bound = [min(s[0][0],s[4][0]), min(s[0][1],s[4][1]), max(s[0][0],s[4][0]), max(s[0][1],s[4][1])] 
				r = (P(s[2])-P(s[0])).mag()
				if self.between_ang(left,st,end): bound[0] = s[2][0]-r 
				if self.between_ang(right,st,end): bound[2] = s[2][0]+r 
				if self.between_ang(top,st,end): bound[3] = s[2][1]+r 
				if self.between_ang(bottom,st,end): bound[1] = s[2][1]-r 
			if bound!=[]:
				bounds = [min(bound[0],bounds[0]), min(bound[1],bounds[1]), max(bound[2],bounds[2]), max(bound[3],bounds[3])]  if bounds!=[] else bound
		return bounds		
				
				
						
		
	def effect(self):
	
		if self.options.function in ['Curve','Area'] :
			if (os.path.isdir(self.options.directory)):
				if (os.path.isfile(self.options.directory+'/header')):
					f = open(self.options.directory+'/header', 'r')
					self.header = f.read()
					f.close()
				else:
					self.header = defaults['header']
				if (os.path.isfile(self.options.directory+'/footer')):
					f = open(self.options.directory+'/footer','r')
					self.footer = f.read()
					f.close()
				else:
					self.footer = defaults['footer']
			else: 
				raise  IOError, 'Directory does not exist!'
				return
		
			gcode = self.header + self.options.unit
			if not self.options.generate_not_parametric_code:
				gcode += """
#4  = %f (Feed)
#5  = %f (Scale x)
#6  = %f (Scale y)
#7  = %f (Scale z)
#8  = %f (Offset x)
#9  = %f (Offset y)
#10 = %f (Offset z)
#11 = %f (Safe distanse)
			""" % (self.options.feed, self.options.Xscale, self.options.Yscale, self.options.Zscale, self.options.Xoffset, self.options.Yoffset, self.options.Zoffset, self.options.Zsafe)

			for id, node in self.selected.iteritems():
				self.biarcGroup = inkex.etree.SubElement( node.getparent(), inkex.addNS('g','svg') )
				break
				
			for id, node in self.selected.iteritems():
				if node.tag == inkex.addNS('path','svg'):
					gcode += '(Found curve %s)\n' % node.get('id').replace('(','').replace(')','').replace('\n','')
					original_d = node.get('d')
		###
		###		Calculating new curve for the path.
		###		
					if self.options.Zstep == 0 : Zstep = 1
				
		###
		###		Compute area cutting if needed
		###

					curve = self.parse_curve(original_d)

#					print_ (curve)
#					k = ''
#					for i in curve: k +=" " +i[1]
#					print_(k)	
#					print_()
					self.draw_curve(curve)
					for step in range( 0,  int(math.ceil( abs( (self.options.Zdepth-self.options.Zsurface)/self.options.Zstep )) ) ):
						Zpos = max(		self.options.Zdepth,		 self.options.Zsurface - abs(self.options.Zstep*(step+1))	)
						gcode += self.generate_gcode(curve,Zpos)

					if self.options.function=='Area' :
						if self.options.tool_diameter<=0 : 
							inkex.errormsg(_("Tool diameter must be > 0!"))				
							return
						curve = self.prepare_curve(curve)	

						bounds = self.get_bounds(curve)
						for cur_num in range(1,self.options.max_area_curves+1):
							d = self.options.tool_diameter*cur_num
							p1 = self.offset_curve(curve, -d)
							p2 = self.offset_curve(curve, d)
							result = self.clip(p1,p2,curve, d)
							bound = self.get_bounds(result)
							self.draw_curve(result, style=biarc_style_dark_area)
							if bound==[] or bound[0]<=bounds[0] or bound[1]<=bounds[1] or bound[2]>=bounds[2] or bound[3]>=bounds[3]:
								break	#	Offset is not smaller than original
							for step in range( 0,  int(math.ceil( abs( (self.options.Zdepth-self.options.Zsurface)/self.options.Zstep )) ) ):
								Zpos = max(		self.options.Zdepth,		 self.options.Zsurface - abs(self.options.Zstep*(step+1))	)
								gcode += self.generate_gcode(result,Zpos)

							bounds = bound

			gcode += self.footer
			try: 	
				f = open(self.options.directory+'/'+self.options.file, "w")	
				f.write(gcode)
				f.close()							
			except: 
				raise  IOError, 'Can not write to specified file!'
				
		elif self.options.function=='Area inkscape' :
			if self.options.tool_diameter<=0 : 
							inkex.errormsg(_("Tool diameter must be > 0!"))				
							return
			for id, node in self.selected.iteritems():
				self.biarcGroup = inkex.etree.SubElement( node.getparent(), inkex.addNS('g','svg') )
				break
									
			for id, node in self.selected.iteritems():
				if node.tag == inkex.addNS('path','svg'):
					d = node.get('d')
					r = self.options.area_inkscape_radius 							
					sign=1 if r>0 else -1
					tool_d = self.options.tool_diameter
					for i in range(self.options.max_area_curves):
						radius = tool_d * (i+0.5) * sign
						if abs(radius)>abs(r): 
							radius = r
						inkex.etree.SubElement(	self.biarcGroup, inkex.addNS('path','svg'), 
										{
											 inkex.addNS('type','sodipodi'):	'inkscape:offset',
											 inkex.addNS('radius','inkscape'):	str(radius),
											 inkex.addNS('original','inkscape'):	d,
											'style':				biarc_style_i['area']
										})
						if radius==r : break 
	
e = Gcode()
e.affect()




